; Common routines for getting addresses etc.

PRESERVE_ALLREGS:
LD (PA_CALLHOLD), HL    ; Preserve HL.
POP HL                  ; Get return address.
LD (PA_RETADDRESS), HL  ; Store as jump location.
LD HL, (PA_CALLHOLD)    ; Restore HL.
PUSH AF                 ; Store main registers.
PUSH BC
PUSH DE
PUSH HL
PUSH IX
PUSH IY
EX AF, AF'              ; Store shadow registers.
PUSH AF
EX AF, AF'
EXX
PUSH BC
PUSH DE
PUSH HL
EXX
DB $C3                  ; First byte of JP ADDRESS instruction.
PA_RETADDRESS:
DW 0                    ; Address to jump to.
PA_CALLHOLD:
DW 0

RESTORE_ALLREGS:
POP HL                  ; Get return address.
LD (RA_RETADDRESS), HL  ; Store as jump location.
EXX                     ; Restore shadow registers.
POP HL
POP DE
POP BC
EXX
EX AF, AF'
POP AF
EX AF, AF'
POP IY                  ; Restore main registers.
POP IX
POP HL
POP DE
POP BC
POP AF
DB $C3                  ; First byte of JP ADDRESS instruction.
RA_RETADDRESS:
DW 0                    ; Address to jump to.

; Returns start address of screen cell (B,C) in DE.
; BC unaffected, A corrupted.
; 63 Ts.
GET_SCR_CELLSTART:
LD A, C                               ; Copy row number.
AND 00011000B                         ; Screen high bits of screen address.
OR 01000000B                          ; Add screen start address.
LD D, A                               ; High byte now set correctly.
LD A, C                               ; Copy row number.
AND 00000111B                         ; Screen low bits of screen address.
RRCA                                  ; Multiply by 32 (row width.)
RRCA
RRCA
ADD A, B                              ; Offset by column (no overflow risk.)
LD E, A                               ; Low byte now set correctly.
RET

; Returns start address of pixel row A in HL.
GET_YPIXPTR:
LD L, A                        ; Preserve copy.
AND 00000111B                  ; Retain only pixel position within cell.
LD H, A                        ; Store position in high byte (*256.)
LD A, L                        ; Restore copy.
RRCA
RRCA
RRCA
AND 00011000B                  ; Retain only screenbank*8.
OR 01000000B                   ; Screen start address high byte ($40.)
OR H                           ; Add pixel position within cell (no overlap.)
LD H, A                        ; High byte now set.
LD A, L                        ; Restore copy.
AND 00111000B                  ; Retain only cell position within bank (*8.)
RLCA                           ; Multiply further by 4 (4*8 = 32.)
RLCA
LD L, A                        ; Low byte now set.
RET

; Returns address and size of item A from index table starting at HL:
; - item address returned in HL
; - item size returned in BC
; - AF is corrupted
; - DE is unchanged
; Index table must:
; - contain no more than 127 entries
; - have a zero-length 128th entry (for 127th entry size calculation)
INDEX_TO_ADDRESS:
PUSH DE                ; Store DE.
SLA A                  ; Multiply item by two (no overflow risk)
LD B, 0
LD C, A                ; BC now contains offset within table.
ADD HL, BC             ; Add offset to table start address.
LD C, (HL)             ; Load BC with data start address.
INC HL
LD B, (HL)
INC HL                 ; Load DE with data end address (start of next item.)
LD E, (HL)
INC HL
LD D, (HL)
EX DE, HL              ; Now HL contains end address.
AND A                  ; Reset Carry flag.
SBC HL, BC             ; Item length = end address - start address.
PUSH HL                ; Swap HL and BC.
LD H, B
LD L, C
POP BC                 ; Now HL contains start address and BC contains size.
POP DE                 ; Restore DE.
RET

; Returns lower byte of global variable number A-128 (where A is >127.)
; If A is not a global variable, it is returned unchanged.
GET_IFGVAR_LOWBYTE:
BIT 7, A
RET Z

  GET_GVAR_LOWBYTE:
  PUSH HL
  LD HL, V_GVAR
  SLA A                ; Multiply A by two, discarding upper bit.
  LD L, A              ; HL now points to correct variable.
  LD A, (HL)           ; Get low byte only.
  POP HL
  RET

; Returns full value of global variable A in DE.
; If A is not a global variable, it is returned unchanged in DE
GET_IFGVAR_FULLBYTE:
LD D, 0
LD E, A
BIT 7, A
RET Z

  GET_GVAR_FULL:
  PUSH HL                   ; Store HL.
  LD HL, V_GVAR             ; 256 byte aligned.
  SLA A                     ; Multiply variable number by two.
  LD L, A                   ; Now HL points to the variable.
  LD E, (HL)
  INC HL
  LD D, (HL)                ; DE now contains the variable's value.
  POP HL                    ; Restore HL.
  RET

; Sets global variable A to DE.
SET_GVAR_FULL:
LD HL, V_GVAR               ; 256 byte aligned.
SLA A                       ; Multiply variable number by two.
LD L, A                     ; Now HL points to the variable.
LD (HL), E
INC HL
LD (HL), D                  ; (HL) now contains DE.
RET

; Gets corresponding global variables for any values above 127 in block.
; Called with block start address in HL, block length in C.
GET_GVAR_BLOCK:
LD A, (HL)
BIT 7, A                    ; Check upper bit.
JR Z, GET_GVAR_BLOCK_JP1
CALL GET_GVAR_LOWBYTE       ; Get lower byte of corresponding global variable.
LD (HL), A                  ; Store in memory.

  GET_GVAR_BLOCK_JP1:
  INC HL                    ; Point to next byte in block.
  DEC C                     ; End of block?
  JR NZ, GET_GVAR_BLOCK
  RET

; Returns in HL address of item A from index table HL.
; Table must have two byte entries, max 128 entries, but can start anywhere
; If BIT 7 of A is set, then item number is taken from variables instead.
GET_ADDRESS_UNALIGNED:
PUSH DE                ; Preserve registers.
BIT 7, A
CALL NZ, GET_GVAR_LOWBYTE
LD E, A
SLA E                  ; Don't need BIT 7 now.
LD D, 0                ; DE now contains offset.
ADD HL, DE             ; HL now points at address.
LD E, (HL)             ; Low byte comes first.
INC HL
LD D, (HL)             ; DE now contains address to return.
EX DE, HL
POP DE                 ; Restore original DE.
RET

; Returns in HL address of item at (D,E) in array of width C, beginning at HL.
GET_ARRAY_ADDRESS:
PUSH DE
PUSH BC
LD B, E
LD E, D
LD D, 0                 ; DE now contains X offset.
ADD HL, DE              ; HL adjusted for X offset.
LD E, B                 ; DE now contains Y offset.
LD A, C
CALL MULT_DE_BY_A_QND
ADD HL, DE              ; HL additionally adjusted for Y offset.
POP BC
POP DE
RET

; Called with XY start in BC, and XY dims in DE.
; Returns clipped dimensions in DE, corrupts BC.
;GET_SCR_CLIPDIMS:
;LD A, 32                       ; Screen width.
;SUB B                          ; X start.
;LD B, A                        ; Screen width - X start.
;LD A, 24                       ; Screen height.
;SUB C                          ; Y start.
;LD C, A                        ; Screen height - Y start.
;                               ; Test width:
;LD A, D                        ; Bitmap width.
;CP B                           ; Screen width - X start.
;JR NC, GET_SCR_CLIPDIMS_JP1
;LD D, B                        ; Clip width.
;
;  GET_SCR_CLIPDIMS_JP1:        ; Test height:
;  LD A, E                      ; Bitmap height.
;  CP C                         ; Screen height - Y start.
;  RET NC
;  LD E, C                      ; Clip height.
;  RET

; Called with HL table start and A entry number (maximum of 127.)
; Returns word from table in IY, corrupts A and HL.
GET_JUMP_FROMTABLE:
SLA A                          ; Each table entry is two bytes long.
ADD A, L                       ; Add A to HL with overflow.
LD L, A
ADC A, H
SUB L
LD H, A
LD A, (HL)                     ; Get jump location.
INC HL
LD H, (HL)
LD L, A
PUSH HL                        ; Copy jump location from HL to IY.
POP IY
RET

; Multiplies DE by A (quick and dirty.)
; This routine does NOT check for overflow, if this is possible call the full routine instead.
; The runtime varies slightly depending on the values in DE and A.
MULT_DE_BY_A_QND:
PUSH HL                ; This routine requires four bytes of stack space.
PUSH BC
LD HL, 0
LD C, 8
LD B, A

  MULT_DE_BY_A_QND_JP1:
  SRL A
  JR NC, MULT_DE_BY_A_QND_JP2
  ADD HL, DE

  MULT_DE_BY_A_QND_JP2:
  DEC C
  JR Z, MULT_DE_BY_A_QND_JP3
  SLA D
  SLA E
  JR NC, MULT_DE_BY_A_QND_JP1
  INC D
  JR MULT_DE_BY_A_QND_JP1

  MULT_DE_BY_A_QND_JP3:
  EX DE, HL
  LD A, B
  POP BC
  POP HL
  RET

; Converts the unsigned integer HL to a character string with leading zeros.
; Preserves all registers.
DOUBLE_TO_STRING:
PUSH AF
PUSH BC
PUSH DE
PUSH HL

LD BC, (DOUBLE_TO_STRING_ADDRESS)
LD DE, 10000           ; Deal with the 10,000s.
SUB A                  ; Set to zero.
CALL DOUBLE_TO_STRING_SUB1
LD DE, 1000            ; Deal with the 1000s.
SUB A                  ; Set to zero.
CALL DOUBLE_TO_STRING_SUB1
LD DE, 100             ; Deal with the 100s.
SUB A                  ; Set to zero.
CALL DOUBLE_TO_STRING_SUB1
LD DE, 10              ; Deal with the 10s (slight overhead as HL now only contains 8-bit value.)
SUB A                  ; Set to zero.
CALL DOUBLE_TO_STRING_SUB1
LD DE, 1               ; Deal with the 1s.
SUB A                  ; Set to zero.
CALL DOUBLE_TO_STRING_SUB1

POP HL
POP DE
POP BC
POP AF
RET

DOUBLE_TO_STRING_SUB1: ; Decimal subtraction and counter loop.
AND A                  ; Reset the Carry flag.
SBC HL, DE
JR C, DOUBLE_TO_STRING_SUB2
INC A
JR DOUBLE_TO_STRING_SUB1

DOUBLE_TO_STRING_SUB2:
ADD HL, DE             ; Reset to value before subtraction caused overflow.
ADD A, 48              ; Offset to convert numerical value to equivalent ASCII character.
LD (BC), A             ; Store result.
INC BC                 ; Get ready for next digit.
RET

DOUBLE_TO_STRING_ADDRESS:
DW MB_PS_NEXTROW

; For use when output characters require modification before printing.
DOUBLE_TO_STRING_BUFFER:
DB 00H
DB 00H
DB 00H
DB 00H
DB 00H
